\\ Copyright 2020, 2021 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the
\\ Free Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but WITHOUT
\\ ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
\\ FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
\\ more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

default(strictargs,1);
default(recover,0);


\\ PARI type t_POLMOD can be used to calculate a single term n in a linear
\\ recurrence.
\\
\\ For example, the Padovan numbers (OEIS A000931),

my(p=Mod(x, x^3 - x - 1)); Padovan(n) = polcoeff(lift(p^n),0);
vector(20,n,n--; Padovan(n)) == \
  [1, 0, 0, 1, 0, 1, 1, 1, 2, 2, 3, 4, 5, 7, 9, 12, 16, 21, 28, 37] || error();

\\ As n increases, polmod powering runs much faster than a simple iteration
\\ of a linear recurrence, and faster and uses less memory than matrix
\\ powering.  However, polmod powering is not so obvious and a little care
\\ is needed when thinking about what values are formed and stepped.
\\
\\ The polmod modulus is a polynomial.  This is the characteristic
\\ polynomial of the recurrence (like recurrence_guess() prints).


\\ The following is a bigger example recurrence with coefficients big enough
\\ and different enough to highlight the places they arise.  (If
\\ coefficients are small or sparse then it's easy to be confused about
\\ which way around they go or which way the signs!)

M = x^4 - 17*x^3 - 13*x^2 + 5*x - 11;      \\ characteristic polynomial
print("M = ",M);

\\ Polmod keeps a polynomial modulo M by subtracting suitable multiples of M
\\ to eliminate terms outside the range x^0 .. x^3.  For example, if have a
\\ high term 5*x^4 then subtract 5*M to eliminate it,

(5*x^4 + x) % M == (5*x^4 + x) - 5*M             || error();
(5*x^4 + x) % M == 85*x^3 + 65*x^2 - 24*x + 55   || error();

\\ For evaluating a linear recurrence, the starting polynomial is x^n.
\\ The coefficients in the polynomial reduced mod M are some recurrence values.
\\
\\ The example M here is the characteristic polynomial of recurrence
\\ 
\\     f(n) - 17*f(n-1) - 13*f(n-2) + 5*f(n-3) - 11*f(n-4) = 0
\\     x^4  - 17*x^3    - 13*x^2    + 5*x      - 11        = M
\\
\\ which is
\\
\\     f(n) = 17*f(n-1) + 13*f(n-2) - 5*f(n-3) + 11*f(n-4)
\\
\\ The following is the matrix powering method for evaluating such a
\\ recurrence starting from given initial values.  The vector contains
\\ consecutive terms [f(n),f(n+1),f(n+2),f(n+3)].  The matrix multiply
\\ shifts them down (rows 0,1,0,0 etc) and sets new f(n+4) at the end
\\ (bottom-most row is the recurrence factors).  In the f() code, "initial"
\\ is a vector of desired initial terms f(0..3).

Mmat = {[ 0,  1,  0,  0;
          0,  0,  1,  0;
          0,  0,  0,  1;
         11, -5, 13, 17 ]};
charpoly(Mmat) == M  || error();

f(n, initial=[0,0,0,1]) = (Mmat^n * initial~)[1];

print("f(n) = ",vector(8,n,n--; f(n)));
vector(8,n,n--; f(n)) == [0, 0, 0, 1, 17, 302, 5350, 94802]     || error();

vector(20,n,n--; f(n) - 17*f(n-1) - 13*f(n-2) + 5*f(n-3) - 11*f(n-4)) == \
vector(20,n,n--; 0)                                             || error();
vector(20,n,n--; f(n)) == \
vector(20,n,n--; 17*f(n-1) + 13*f(n-2) - 5*f(n-3) + 11*f(n-4))  || error();


\\ The polmod method for the same thing starts with polynomial x and powers
\\ it up as a Mod(x,M).  f_by_polmod() parameter "term" is which term to
\\ extract from the reduced polynomial.  The default 3 (the highest term)
\\ corresponds to the default "initial" of f(n).
\\
\\ f_by_polmod() has a lift() because polcoeff() wants type t_POL rather
\\ than t_POLMOD.  A distinction between a polynomial and a polynomial
\\ residue class is more important to mathematicians than programmers, but
\\ in any case the lift is needed.
\\
PM = Mod(x, M);
f_by_polmod(n, term=3) = polcoeff(lift(PM^n), term);

vector(20,n,n--; f(n)) == vector(20,n,n--; f_by_polmod(n))  || error();

\\ The same x powering can be done with a plain polynomial x^n created and
\\ then reduced mod M,
\\
vector(20,n,n--; f(n)) == \
vector(20,n,n--; polcoeff(x^n % M, 3)) || error();

\\ But polmod is much faster than x^n because polmod uses PARI's powering
\\ algorithms (successive squaring and stepping, maybe more one day),
\\ instead of a general-purpose polynomial remainder.
\\
\\ Polmod is faster than matrix multiplies too, essentially since polmod
\\ calculates and keeps 4 quantities instead of the matrix 16 quantities.
\\ The difference becomes yet bigger when the recurrence order is bigger.
{
  \\ Here's some example speed measures.  f_by_polmod() is fastest, and then
  \\ matrix f() is faster than plain polynomial % M reduction.  Change to
  \\ "if(1," to enable this code, and adjust "reps" to see enough difference
  \\ according as your computer is fast or slow.
  if(0,
     print(); gettime();
     my(n=1000);
     for(rep=1,600, f_by_polmod(n)); print("f_by_polmod() time ",gettime());
     for(rep=1,600, f(n));           print("f() matrix time    ",gettime());
     for(rep=1,600, x^n % M);        print("plain %M time      ",gettime());
    );
}

\\ As noted above, the coefficient of x^3 in the polmod is f(n) of initial
\\ values 0,0,0,1.  Each of the coefficients are
\\
\\             initial values    coeff of
\\             --------------    --------
\\     a(n) =     1,0,0,0          x^0   (ie. constant term)
\\     b(n) =     0,1,0,0          x^1
\\     c(n) =     0,0,1,0          x^2
\\     d(n) =     0,0,0,1          x^3

a(n) = f(n, [1,0,0,0]);
b(n) = f(n, [0,1,0,0]);
c(n) = f(n, [0,0,1,0]);
d(n) = f(n, [0,0,0,1]);

vector(20,n,n--; a(n)) == vector(20,n,n--; f_by_polmod(n, 0))  || error();
vector(20,n,n--; b(n)) == vector(20,n,n--; f_by_polmod(n, 1))  || error();
vector(20,n,n--; c(n)) == vector(20,n,n--; f_by_polmod(n, 2))  || error();
vector(20,n,n--; d(n)) == vector(20,n,n--; f_by_polmod(n, 3))  || error();

\\ As initial values, these are clear simply since x^n for n=0..3 is
\\ unchanged by mod M, so coefficient 1 or 0 at these places.
\\
\\ To see that the coefficients all follow the recurrence, consider what the
\\ reduction mod M does to a new factor x.  Suppose PM^n is
\\
\\     PM^n = a(n) + b(n)*x + c(n)*x^2 + d(n)*x^3

vector(20,n,n--; lift(PM^n)) == \
vector(20,n,n--; a(n) + b(n)*x + c(n)*x^2 + d(n)*x^3) || error();

\\ Then multiplying by x for PM^(n+1) is reduced mod M by subtracting d(n)*M,
\\
\\     PM^(n+1)  = x*PM^n
\\              == x*PM^n - d(n)*M     mod M
{
  vector(20,n,n--; lift(PM^(n+1))) == \
  vector(20,n,n--; x*lift(PM^n) - d(n)*M)      || error();

  vector(20,n,n--; lift(PM^(n+1))) == \
  vector(20,n,n--;   (     + 11*d(n))
                   + (a(n) -  5*d(n)) *x
                   + (b(n) + 13*d(n)) *x^2
                   + (c(n) + 17*d(n)) *x^3
                   + (d(n) -    d(n)) *x^4 )   || error();
}

\\ These new coefficients are a(n+1) through d(n+1),
\\
vector(20,n,n--; a(n+1)) == vector(20,n,n--;        11*d(n))   || error();
vector(20,n,n--; b(n+1)) == vector(20,n,n--; a(n) -  5*d(n))   || error();
vector(20,n,n--; c(n+1)) == vector(20,n,n--; b(n) + 13*d(n))   || error();
vector(20,n,n--; d(n+1)) == vector(20,n,n--; c(n) + 17*d(n))   || error();

\\ For c, the c initial values plus the initial values of 17*d are exactly
\\ d(1..4),
\\
\\      c(0..3) =     0, 0, 1, 0
\\   17*d(0..3) =     0, 0, 0, 17    add
\\                 --------------
\\    d(0,1..4) =  0, 0, 0, 1, 17    gives
\\
\\ And similarly the others, each due to the way the position of the 1 in
\\ the initial values selects the respective recurrence factor.

vector(4,n,n--;    c(n)) == [   0, 0, 1, 0]       || error();
vector(4,n,n--; 17*d(n)) == [   0, 0, 0, 17]      || error();
vector(5,n,n--;    d(n)) == [0, 0, 0, 1, 17]      || error();

vector(4,n,n--;    b(n)) == [   0, 1, 0, 0]       || error();
vector(4,n,n--; 13*d(n)) == [   0, 0, 0, 13]      || error();
vector(5,n,n--;    c(n)) == [0, 0, 1, 0, 13]      || error();

vector(4,n,n--;    a(n)) == [   1, 0, 0, 0]       || error();
vector(4,n,n--; -5*d(n)) == [   0, 0, 0, -5]      || error();
vector(5,n,n--;    b(n)) == [0, 1, 0, 0, -5]      || error();

vector(4,n,n--; 11*d(n)) == [   0, 0, 0, 11]      || error();
vector(5,n,n--;    a(n)) == [1, 0, 0, 0, 11]      || error();


\\-----------------------------------------------------------------------------
\\ Generating Functions
\\
\\ The generating function for coefficient d(n) is x^3 / RM where denominator
\\ RM is the reciprocal (terms reversed) of the characteristic polynomial M.
\\ This is the usual way to make a generating function for initial values
\\ 0,0,0,1.
\\
\\ The others follow from the relations above such as
\\
\\     d(n+1) = c(n) + 17*d(n)
\\     c(n) = d(n+1) - 17*d(n)
\\
\\ which brings the recurrence factors into the generating function
\\ numerators.
\\
\\ It's often not easy to notice by eye when the numerator is a form like
\\ this, and thus one of the polmod coefficients.  If there's some 0s among
\\ the recurrence factors then that can make two of the polmod coefficients
\\ the same sequence, just different offset.  For example if 13 was instead
\\ 0 then gb(x) and gc(x) would differ only in a factor of x (an index
\\ shift).

RM = polrecip(M);
RM == 1 - 17*x - 13*x^2 + 5*x^3 - 11*x^4     || error();

ga(x) = ( 1   - 17*x  - 13*x^2 +  5*x^3 ) / RM;
gb(x) = (         x   - 17*x^2 - 13*x^3 ) / RM;
gc(x) = (                  x^2 - 17*x^3 ) / RM;
gd(x) = (                           x^3 ) / RM;

ga(x) + O(x^6) == 1 + 0*x + 0*x^2 + 0*x^3 + 11*x^4 + 187*x^5   || error();
gb(x) + O(x^6) == 0 + 1*x + 0*x^2 + 0*x^3 -  5*x^4 -  74*x^5   || error();
gc(x) + O(x^6) == 0 + 0*x + 1*x^2 + 0*x^3 + 13*x^4 + 216*x^5   || error();
gd(x) + O(x^6) == 0 + 0*x + 0*x^2 + 1*x^3 + 17*x^4 + 302*x^5   || error();

\\ g is a t_RFRAC.  Return a vector of len many coeffs starting from x^0.
gf_to_terms(g,len) = g+=O(x^len); vector(len,i,i--; polcoeff(g,i));

gf_to_terms(ga(x),20) == vector(20,n,n--; a(n))   || error();
gf_to_terms(gb(x),20) == vector(20,n,n--; b(n))   || error();
gf_to_terms(gc(x),20) == vector(20,n,n--; c(n))   || error();
gf_to_terms(gd(x),20) == vector(20,n,n--; d(n))   || error();


\\-----------------------------------------------------------------------------
\\ Initial Terms 1,1,1,1
\\
\\ The polmod coeffs can be summed to give a recurrence with initial values
\\ 1,1,1,1.  This is good value for money from the powering, in the sense
\\ that all the coeffs so carefully made by squarings and multiplyings are
\\ used (with relatively cheap additions).
\\
\\ vecsum(Vec()) is so convenient that it's worth checking whether 1,1,1,1
\\ occurs in a given recurrence somewhere earlier or later than an actual
\\ desired start, and then apply a suitable offset n-1 or whatever power so
\\ as to go relative to there.
  
vector(20,n,n--; f(n, [1,1,1,1])) == \
vector(20,n,n--; vecsum(Vec(lift(PM^n))))   || error();


\\ Vec(poly) gives the poly coefficients from the highest non-0 downward.
\\ A truncated Vec call like Vec(etc,2) gives the highest and the
\\ immediately following (which might be 0).  If you know the coeff of x^3
\\ is not 0, then a truncated Vec can select say x^3 and x2 for initial
\\ values 0,0,1,1.  Be careful that x^3 will be non-0 though, otherwise the
\\ result is say x^2 and x^1 which will be quite different (in general).

\\ Note the following is only for n>=2:
vector(20,n,n++; f(n, [0,0,1,1])) == \
vector(20,n,n++; vecsum(Vec(lift(PM^n),2))) || error();

\\ Any constant factor throughout like initial values 3,3,3,3 are as easy as
\\ 3*vecsum(...).

\\ Other compact expressions can suit other initial values.  A geometric
\\ progression like 1,2,4,8 can use lift then subst to let the polynomial
\\ powers x^0..x^3 give exactly those geometric powers 2^0..2^3,
\\
vector(20,n,n--; f(n, [1,2,4,8])) == \
vector(20,n,n--; subst(lift(PM^n),x,2)) || error();
\\
\\ Or initial values like 0,1,2,4 could omit the first term with a
\\ round-down polynomial shift
\\
vector(20,n,n--; f(n, [0,1,2,4])) == \
vector(20,n,n--; subst(lift(PM^n)\x, x, 2)) || error();

\\ Again any constant factor like initial values 1/2,1,2,4 are as easy as
\\ (1/2)*vecsum(...).
\\
vector(20,n,n--; f(n, [1/2, 1, 2, 4])) == \
vector(20,n,n--; subst(lift(PM^n),x,2)/2) || error();

\\ If working with integers then this sort of initial term 1/2 might exist
\\ only by stepping the recurrence backwards by one place to see a term
\\ preceding the actual desired initial terms.  (Both polmod powering and
\\ matrix powering allow negative "n" for calculating earlier terms.  See
\\ below for more on that.)


\\-----------------------------------------------------------------------------
\\ Polynomial Multiplier
\\
\\ When one of the special cases above doesn't apply, a general approach for
\\ a combination of coefficients from the polmod is to multiply by a
\\ polynomial while still within the modulo M.
\\
\\ In terms of a generating function, such a multiplier can be the
\\ polrecip() of the numerator in the gf.  The high polcoeff() of the
\\ resulting polmod is then the desired terms.  This is since, as described
\\ above, gf x^3/RM is the plain polcoeff(lift(PM^n),3) and a multiplier
\\ x*PM^n brings up an earlier term, so gf x^2/RM, etc.

gf_to_terms(x^3/RM,20)==vector(20,n,n--;polcoeff(lift(1  *PM^n),3)) ||error();
gf_to_terms(x^2/RM,20)==vector(20,n,n--;polcoeff(lift(x  *PM^n),3)) ||error();
gf_to_terms(x  /RM,20)==vector(20,n,n--;polcoeff(lift(x^2*PM^n),3)) ||error();
gf_to_terms(1  /RM,20)==vector(20,n,n--;polcoeff(lift(x^3*PM^n),3)) ||error();
{
  my(gnumerator = 97     + 123*x   + 54*x^2 + 71*x^3,
     g          = gnumerator/RM,
     m          = 97*x^3 + 123*x^2 + 54*x   + 71);
  m == polrecip(gnumerator)                        || error();
  gf_to_terms(g,20) == vector(20,n,n--; polcoeff(lift(m*PM^n),3))  || error();
}


\\-----------------------------------------------------------------------------
\\ Negative Powers

\\ The polmod method and the matrix method both allow negative n.
\\ Negative n means run the recurrence backwards to earlier terms.

vector(20,n,n--; f(-n)) == vector(20,n,n--; f_by_polmod(-n))  || error();

\\ In the polmod, such a reversal can also be done by starting from x^-1,
\\ since Mod(x^-1,M) == Mod(x,M)^-1.

vector(20,n,n--; f(-n)) == \
vector(20,n,n--; polcoeff(lift(Mod(x^-1,M) ^ n), 3))          || error();

\\ However, such a Mod(x^-1,M) is reduced to a polynomial with degree <= 3
\\ and no negative terms,

lift(Mod(x^-1,M)) == 1/11*x^3 - 17/11*x^2 - 13/11*x + 5/11    || error();

\\ Written out explicitly like this, such a starting point for powering is
\\ less obvious than a Mod(x,M), but it works.
\\
\\ In general, reversed terms from a particular recurrence are probably
\\ cleaner by forming a new recurrence with suitable new charpoly
\\ (polrecip() of the original) and starting at Mod(x,M).
\\
\\ In the following, notice the polcoeff is term 0 since f initial values
\\ 0,0,0,1 become 1,0,0,0 when read in reverse.  Index n is aligned in the
\\ two forms by offset "n+3".

M_recip = polrecip(M);
vector(20,n,n--; f(-n)) == \
vector(20,n,n--; polcoeff(lift(Mod(x,M_recip) ^(n+3)), 0))    || error();


\\-----------------------------------------------------------------------------
\\ Squared x Powering

\\ It might be wondered whether the polmod could power up an x^2.  Yes it
\\ can, and doing so is every second term of the recurrence, since

vector(20,n,n--; Mod(x^2,M)^  n ) == \
vector(20,n,n--; Mod(x,  M)^(2*n))                   || error();

vector(20,n,n--; f(2*n)) == \
vector(20,n,n--; polcoeff(lift(Mod(x^2,M)^n), 3))    || error();

\\ Every second term (or any k'th terms in general) like this is also a
\\ linear recurrence, but one not necessarily starting 0,0,0,1 and so
\\ possibly needing some factors.  lindep() is an easy way to find them,
\\ knowing for example that an order 4 recurrence with factors correct on
\\ 4 terms is certain to be correct everywhere.
\\
M2 = x^4 - 315*x^3 + 317*x^2 + 261*x + 121;
{
  lindep([vector(20,n,n--; polcoeff(lift(Mod(x,M2)^n), 0)),
          vector(20,n,n--; polcoeff(lift(Mod(x,M2)^n), 1)),
          vector(20,n,n--; polcoeff(lift(Mod(x,M2)^n), 2)),
          vector(20,n,n--; polcoeff(lift(Mod(x,M2)^n), 3)),
          vector(20,n,n--; f(2*n))])
  ==
  [0, 0, -17, -5350, 1]~        || error();

  vector(20,n,n--; f(2*n)) ==
  vector(20,n,n--; my(p=lift(Mod(x,M2)^n));
                   17*polcoeff(p,2) + 5350*polcoeff(p,3))
  || error();
}


\\-----------------------------------------------------------------------------
\\ Constant Folding

\\ As a note on final polished code, GP circa its version 2.11.x does no
\\ constant folding.  So if a polynomial is written out x^3 - 17*x^2 - ...
\\ then those multiplies and adds run as written every time.
\\
\\ If a constant is big and only wanted occasionally, then making it when
\\ needed is good.  If small then pre-calculating and holding it outside the
\\ function will be both smaller and faster.  Pre-calculating when big would
\\ gain run-time speed at the cost of space and perhaps startup time.
\\
\\ In code for a library or other re-use, write x as symbol 'x to be safe
\\ against any global variable x the user might have.
\\
\\ A local lexical variable like the following is visible to the function
\\ and nowhere else.  A few functions can be created in the one statement
\\ sharing the same lexical (with enough parens).  Or a global (with
\\ suitably unique name) works equally well if it's wanted in widely
\\ spread-out places.

my(p=Mod('x, 'x^4 - 17*'x^3 - 13*'x^2 + 5*'x - 11)); \
  f_polished(n) = polcoeff(lift(p^n),3);

vector(20,n,n--; f(n)) == vector(20,n,n--; f_polished(n))  || error();


\\-----------------------------------------------------------------------------
print("end");
